//------------------------------------------------------------------------------
// "DADA-poem"
//
// - a sound poem generator in dadaist fashion using verses without words
//
// desc: Dadaistic poems have a simultaneously playful and nihilistic spirit, 
//       A subcaterogy called "Simultangedicht" "simulatenous poem" is a form
//       where non-existent words are used to create a phonetic experience.
//       One such poem is "Karawane" by Hugo Ball from 1916. 
//       (https://www.aklein.vsfd.hoodmuseum.dartmouth.edu/karawane,
//       http://members.peak.org/~dadaist/Deutsch/Graphiken/karawane.html)

// NOTE: this needs to run from command line chuck
//       (due to ConsoleInput)
//       > chuck dada-poem-generator.ck

// this file is based on example code by Ge Wang: starter-prompt.ck,
// poem-randomwalk.ck, poem-i-feel.ck, poem-spew.ck

// author: Laura Schuetz
//------------------------------------------------------------------------------

// make a ConsoleInput
ConsoleInput in;
// tokenizer
StringTokenizer tok;
// line
string line[0];

// our default model
Word2Vec model;

// loading any default here
"glove-wiki-gigaword-50-tsne-2.txt" => string filepath;
<<< "loading 2D model:", filepath, "..." >>>;
model.load( filepath );
//print( model );

// words and vectors
"dada" => string cur_word;
"da" => string cur_stem;
"" => string cur_sim_word; // similar word from model
// word vector
float word_vec[model.dim()];
// number of nearest words to retrieve for each word
25 => int K_NEAREST;
// search results
string words[K_NEAREST];

// ranges for each dimension (for sound mapping)
float mins[0], maxs[0];
// get bounds for each dimension
model.minMax( mins, maxs );

// sound
FMVoices vocFM => JCRev rev => Echo b => dac;
220.0 => vocFM.freq;
0.95 => vocFM.gain;
.8 => rev.gain;
.2 => rev.mix;
100::ms => b.max;
100::ms => b.delay;
.50 => b.mix;

// sound background
Shakers shake => NRev reverb => Chorus chor => dac;
.1 => reverb.mix;

// timing
200::ms => dur T_LINE_PAUSE; // pause after each line
100::ms => dur T_WORD; // duration per word

// spork to run a function in parallel
spork ~ shakers();

// some background
fun void shakers()
{
    while( true )
    {
        // set pitch
        48 => Std.mtof => shake.freq;
        .5 => shake.noteOn;
        // advance time
        800::ms => now;
        1 => shake.noteOff;
    }
}

//welcome message and instructions
cherr <= IO.newline();
<<< "This is a Dadaist sound poetry generator. 
Type words to prompt a poem. 
If you are done, type 'bye'.", "" >>>;
cherr <= IO.newline();

while( true )
{
    in.prompt( "type a few words >" ) => now;
    
    // read
    while( in.more() )
    {
        // clear tokens array
        line.clear();
        // get the input line as one string; send to tokenizer
        tok.set( in.getLine() );
        // tokenize line into words separated by white space
        while( tok.more() )
        {
            // put into array (lowercased)
            line << tok.next().lower();
        }
        
        // if non-empty
        if( line.size() )
        {
            if( line[0] == "exit" || line[0] == "quit" || line[0] == "bye")
            {
                // print
                cherr <= IO.newline();
                <<< "Dada! Dada!", "" >>>;
                cherr <= IO.newline();
                me.exit();
            }
            // dada-fy the word and play sound
            dada( line );
        }
    }
}

// do something with the text input 
fun void dada( string input[] )
{
    // new line
    cherr <= IO.newline();
    
    // loop over string array entries to sonify the words
    for( int i; i < input.size(); i++ )
    {
        input[i] => cur_word;
        // get word vector (for sonification)
        model.getVector( cur_word, word_vec );
        // set vowel for line
        Math.random2f(.3, 1) => vocFM.vowel;
        // stem, print and sonify word
        stem( cur_word ); 
        print( cur_stem ); 
        sonify( cur_stem ); 
        wait( T_WORD );
        
        // get similar words and randomly choose one
        model.getSimilar( cur_word, words.size(), words );
        
        // get multiple similar words depending on word length
        for( int j; j < cur_word.length(); j++)
        {
            words[Math.random2(0, words.size()-1)] => cur_sim_word;
            // stem word
            stem( cur_sim_word );
            
            // print without space if even number
            if (cur_word.length() % 2 == 0) 
            {
                print( cur_stem );
            } 
            else 
            {
                print_space( cur_stem );
            }
            // sonify similar word
            sonify( cur_stem );
            wait( T_LINE_PAUSE ); 
        }
        // new line
        cherr <= IO.newline();
    }
    // new line
    cherr <= IO.newline();
}

//----------------- HELPER FUNCTIONS -------------------//

// print a word with space after
fun void stem( string word )
{
    if (word.length() > 3)
    {
        Math.random2(2,3) => int range;
        word.substring(0, range) => cur_stem;
        
    }
    else if (word.length() <= 3) 
    {
        word => cur_stem;
    }
}

// print a word with space after
fun void print( string word )
{
    // print it
    chout <= word; chout.flush();
}

// print a word with space after
fun void print_space( string word )
{
    // print it
    chout <= word <= " "; chout.flush();
}

// sonify
fun void sonify( string word )
{
    Math.random2(24, 60) => Std.mtof => vocFM.freq;
    Math.random2f(.6,.8) => vocFM.noteOn;
    T_WORD => now;
    1 => vocFM.noteOff;
}

// wait
fun void wait( dur T )
{    
    // let time pass, let sound...sound
    T => now;
}
